<!DOCTYPE html>
<html lang="en-us">

<head>
  <meta charset="utf-8">
  <meta name="viewport" content="width=device-width, initial-scale=1.0">
  <title>streaQ | {{ title }}</title>
  <!-- Favicon -->
  <link rel="icon" type="image/x-icon" href="https://www.tastyware.dev/static/favicon.ico">
  <link href="https://cdn.jsdelivr.net/npm/bootstrap@5.3.3/dist/css/bootstrap.min.css" rel="stylesheet"
    integrity="sha384-QWTKZyjpPEjISv5WaRU9OFeRpok6YctnYmDr5pNlyT2bRjXh0JMhjY6hW+ALEwIH" crossorigin="anonymous">
  <script src="https://unpkg.com/htmx.org@2.0.4"></script>
</head>

<body>
  <nav class="navbar navbar-expand-sm bg-body-tertiary">
    <div class="container">
      <a class="navbar-brand fw-bold" href="{{ base_url }}queue">streaQ</a>
        <ul class="navbar-nav">
          <li class="nav-item">
            <a class="nav-link {% block queue %}{% endblock %}" href="{{ base_url }}queue">Queue</a>
          </li>
          <li class="nav-item">
            <a class="nav-link {% block workers %}{% endblock %}" href="{{ base_url }}workers">Workers</a>
          </li>
          <li class="nav-item">
            <a class="nav-link {% block cron %}{% endblock %}" href="{{ base_url }}cron">Cron Jobs</a>
          </li>
        </ul>
      </div>
    </nav>
  {% block content %}
  {% endblock %}
  <script src="https://cdn.jsdelivr.net/npm/bootstrap@5.3.3/dist/js/bootstrap.bundle.min.js"
    integrity="sha384-YvpcrYf0tY3lHB60NNkmXc5s9fDVZLESaAA55NDzOxhy9GkcIdslK1eN7N6jIeHz"
    crossorigin="anonymous"></script>
</body>

</html>
