REDIS_ABORT = ":aborted"
REDIS_CHANNEL = ":channels:"
REDIS_CRON = ":cron:"
REDIS_DEPENDENCIES = ":task:dependencies:"
REDIS_DEPENDENTS = ":task:dependents:"
REDIS_GROUP = "workers"
REDIS_HEALTH = ":health"
REDIS_PREFIX = "streaq:"
REDIS_PREVIOUS = ":task:previous:"
REDIS_RESULT = ":task:results:"
REDIS_RETRY = ":task:retry:"
REDIS_RUNNING = ":task:running:"
REDIS_QUEUE = ":queues:delayed:"
REDIS_STREAM = ":queues:"
REDIS_TASK = ":task:data:"
REDIS_TIMEOUT = ":queues:timeout:"
REDIS_UNIQUE = ":task:lock:"
